#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Spatial_Acceleration/Nearest_Neighbors.h>

#ifdef USE_TORCH
#include <torch/script.h>
#include <torch/torch.h>
#endif

using namespace PhysBAM;
typedef double T;
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;
typedef VECTOR<T,6> SV;

#ifdef USE_TORCH
int prec=6;

torch::Device device(torch::kCPU);
T Poly6Kernel(T h,T x)
{
    if(x>=h) return 0;
    T c=(T)315/(64*pi*pow<9>(h));  // not use pow 
    T hx=h*h-x*x,hx2=c*hx*hx;
    return hx*hx2;
}

void Compute_Features(GRID<TV>& grid,const ARRAY<TV>& X,ARRAY<T,IV> &m0,T m0_r)
{
    T dx=grid.dX.Min();
    T h=m0_r*dx;
    
    T d_z=Poly6Kernel(h,0);
    ARRAY<T> densitys(X.m,init_new,d_z);

    Nearest_Neighbors<TV,T>(X,
        [h,&densitys](int i,int j,T d)
        {
            T d_v=Poly6Kernel(h,d);
            densitys(i)+=d_v;
            densitys(j)+=d_v;
        },h);

    for(int i=0;i<X.m;i++)
    {
        RANGE<IV> neighbors;
        neighbors=m0.Domain_Indices().Intersect
            (RANGE<IV>(grid.Cell_I(X(i)-h),grid.Cell_I(X(i)+h)+1));
        
        for(RANGE_ITERATOR<TV::m> it(neighbors);it.Valid();it.Next())
        {
            TV dist=grid.Cell_X(it.index)-X(i);
            T m=dist.Magnitude_Squared();
            T m0_v=Poly6Kernel(h,sqrt(m));
            m0(it.index)+=m0_v/densitys(i);
        }
    }
}

void Print(std::ostream& o,T x,char c=' ')
{
    LOG::fprintf(o,"%.*f%c",prec,x,c);
}

template<class T,int d>
void Print(std::ostream& o,const VECTOR<T,d>& x,char c=' ')
{
    for(int i=0;i<d;i++)
        Print(o,x(i),i<d-1?' ':c);
}

void Construct_Features(const GRID<TV>& grid,
    const ARRAY<T,IV>& m0,ARRAY<bool,IV>& i_flags,
    ARRAY<bool,IV>& o_flags,RANDOM_NUMBERS<T>& rand,
    int quota,ARRAY<T,IV>& node_phi,const std::string& prefix,
    int network_width)
{
    T dx=grid.dX.Min();
    std::ofstream m0_stream(prefix+".m0");
        
    std::ofstream depth_stream;    
    depth_stream.open(prefix+".depth");
    
    ARRAY<IV> node_indices;
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices());it.Valid();it.Next())
    {
        if(o_flags(it.index) || i_flags(it.index)) continue;
        node_indices.Append(it.index);
    }    

    rand.Random_Shuffle(node_indices);
    int inside=0, outside=0;
    ARRAY<IV> i_indices, o_indices;
     
    for(auto p=node_indices.begin();p!=node_indices.end();p++){
        if(inside==quota && outside==quota) break;
            
        T depth=node_phi(*p)/dx;
        if(depth<0 && !i_flags(*p) && depth>-2) {
            if(inside==quota) continue;
            else { inside++; i_indices.Append(*p);}
        }
        if(depth>0 && !o_flags(*p) && depth<2)
        {
            if(outside==quota) continue;
            else {outside++; o_indices.Append(*p);}
        }
    }
            
    if(inside<outside) o_indices.Resize(inside);
    if(inside>outside) i_indices.Resize(outside);
    i_indices.Append(o_indices);
    node_indices=i_indices;

    for(auto p=node_indices.begin();p!=node_indices.end();p++)
    {
        RANGE<IV> depth_grid(*p-1,*p+2);
        bool abnormal=false;
        for(RANGE_ITERATOR<TV::m> ir(depth_grid);ir.Valid();ir.Next()){
            if(!node_phi.Valid_Index(ir.index))
            {abnormal=true;break;}
            if(std::abs(node_phi(ir.index)/dx)>3)
            {abnormal=true;break;}
        }
        if(abnormal) continue;

        RANGE<IV> cubic_grid(*p-network_width,*p+network_width);
        for(RANGE_ITERATOR<TV::m> ir(cubic_grid);ir.Valid();ir.Next())
            Print(m0_stream,m0(ir.index));

        for(RANGE_ITERATOR<TV::m> ir(depth_grid);ir.Valid();ir.Next())
            Print(depth_stream,node_phi(ir.index)/dx);
            
        m0_stream<<std::endl;
        depth_stream<<std::endl;
    }
}



void Find_Nodes_Inside_Surface(const GRID<TV>& grid,const ARRAY<TV>& X,int acc_radius,int threshold,ARRAY<bool,IV> &i_flags,ARRAY<bool,IV> &o_flags)
{
    ARRAY<int,IV> p_flags(grid.Cell_Indices(acc_radius*2));
    for(int i=0;i<X.m;i++)
    {
        auto cell=grid.Cell_I(X(i));
        if(p_flags.Valid_Index(cell))
            p_flags(cell)=1;
    }

    ARRAY<int,IV> table(grid.Node_Indices(acc_radius*2));
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices(acc_radius));it.Valid();it.Next())
    {
        int x=it.index.x;int y=it.index.y; int z=it.index.z;
        table(it.index)=table(IV(x-1,y,z))+table(IV(x,y-1,z))+table(IV(x,y,z-1))
            -table(IV(x-1,y-1,z))-table(IV(x,y-1,z-1))-table(IV(x-1,y,z-1))
            +table(it.index-1)+p_flags(it.index-1);
    }

    int sum_i=0,sum_o=0;
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices());it.Valid();it.Next())
    {
        IV pos=it.index+acc_radius;
        int x=pos.x;int y=pos.y;int z=pos.z;
        int diameter=2*acc_radius;
        // determine inside nodes
        int neighbors=table(pos)-table(IV(x-diameter,y,z))
            -table(IV(x,y-diameter,z))-table(IV(x,y,z-diameter))
            +table(IV(x-diameter,y-diameter,z))+table(IV(x,y-diameter,z-diameter))
            +table(IV(x-diameter,y,z-diameter))-table(pos-diameter);

        if(neighbors>threshold)
        {
            i_flags(it.index)=true;
            sum_i++;
            continue;
        }
        //determine outside nodes
        pos=it.index+acc_radius/2;
        x=pos.x;y=pos.y;z=pos.z;
        neighbors=table(pos)-table(IV(x-acc_radius,y,z))
            -table(IV(x,y-acc_radius,z))-table(IV(x,y,z-acc_radius))
            +table(IV(x-acc_radius,y-acc_radius,z))+table(IV(x,y-acc_radius,z-acc_radius))
            +table(IV(x-acc_radius,y,z-acc_radius))-table(pos-acc_radius);
        
        if(neighbors==0)
        {
            o_flags(it.index)=true;
            sum_o++;
        }
    }
}

int main(int argc,char *argv[])
{
    int seed=123;
    int network_width=5;
    std::string file_xyz,file_grid,file_node_phi;
    int quota=0;
    std::string prefix;
    int threshold=410; // node is inside if cells with particles>threshold
    int acc_radius=4; // acceleration radius
    T m0_r=4;
    
    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-seed",&seed,"seed","Random number seed");
    parse_args.Add("-network_width",&network_width,"width","Radius of cells for network");
    parse_args.Add("-m0_r",&m0_r,"radius","Radius of m0 feature");
    parse_args.Add("-quota",&quota,"quota","Accept inside and outside features of quota size, use 0 when testing");
    parse_args.Add("-i_xyz",&file_xyz,"file","Input xyz positions file");
    parse_args.Add("-i_grid",&file_grid,"file","Input grid file");
    parse_args.Add("-i_phi",&file_node_phi,"file","Input node phi file (optional)");
    parse_args.Add("-o_prefix",&prefix,"file prefix","Prefix of output file");
    parse_args.Add("-prec",&prec,"precision","feature precision");
    parse_args.Parse();
   
    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(seed);
    ARRAY<TV> X;
    ARRAY<T,IV> node_phi;
    Read_From_File(file_xyz,X);
    GRID<TV> grid;
    if(file_grid.size())
        Read_From_File(file_grid,grid);

    if(file_node_phi.size())
        Read_From_File(file_node_phi,node_phi);
    
    // acceleration
    ARRAY<bool,IV> i_flags(grid.Node_Indices());
    ARRAY<bool,IV> o_flags(grid.Node_Indices());
    Find_Nodes_Inside_Surface(grid,X,acc_radius,threshold,i_flags,o_flags);

    ARRAY<T,IV> m0(grid.Cell_Indices(2*(network_width)));
    Compute_Features(grid,X,m0,m0_r);
    
    Construct_Features(grid,m0,i_flags,o_flags,rand,quota,node_phi,prefix,network_width);
}
#else
int main(int argc,char *argv[])
{
    PHYSBAM_FATAL_ERROR("You must enable USE_TORCH to use this.");
    return 0;
}
#endif
