import os
import torch
import torch.utils.data as data
import numpy as np
import random

torch.set_default_dtype(torch.float64) 

class DataSampler(data.sampler.Sampler):
    def __init__(self, data_source, shuffle_data=True):
        self.data_source = data_source
        self.total_patch_count = list(range(len(data_source)))        
        if(shuffle_data): random.shuffle(self.total_patch_count)

    def __iter__(self):
        return iter(self.total_patch_count)

    def __len__(self):
        return len(self.total_patch_count)
 

class Shape():
    def __init__(self, feat=None, depth=None):
        self.feat = feat
        self.depth = depth


class Cache():
    def __init__(self, capacity, loader, loadfunc):
        self.elements = {}
        self.used_at = {}
        self.capacity = capacity
        self.loader = loader
        self.loadfunc = loadfunc
        self.counter = 0

    def get(self, element_id):
        if element_id not in self.elements:
            # cache miss
            # if at capacity, throw out least recently used item
            if len(self.elements) >= self.capacity:
                remove_id = min(self.used_at, key=self.used_at.get)
                del self.elements[remove_id]
                del self.used_at[remove_id]

            # load element
            self.elements[element_id] = self.loadfunc(self.loader, element_id)

        self.used_at[element_id] = self.counter
        self.counter += 1
        return self.elements[element_id]


class PatchDataset(data.Dataset):
    def __init__(self, root, shape_names, cache_capacity=150):
        # initialize parameters
        self.root = root
        self.shape_cache = Cache(cache_capacity, self, PatchDataset.load_shape_by_index)
        self.shape_names = shape_names
        self.shape_patch_count = []
        
        # get basic information for each shape in the dataset
        for shape_ind, shape_name in enumerate(self.shape_names):
            print('getting information for shape %s' % (shape_name))
            shape = self.shape_cache.get(shape_ind)
            self.shape_patch_count.append(shape.feat.shape[0])            

    # returns a patch centered at the point with the given global index
    def __getitem__(self, index):
        # find shape that contains the point with given global index
        shape_ind, patch_ind = self.shape_index(index)
        shape = self.shape_cache.get(shape_ind)
        feat_patch = shape.feat[patch_ind] 
        depth_patch = shape.depth[patch_ind]         
        return (feat_patch, depth_patch)

    # return the sum of patch count for all shapes
    def __len__(self):
        return sum(self.shape_patch_count)

    # translate global (dataset-wide) point index to shape index & local (shape-wide) point index
    def shape_index(self, index):
        shape_patch_offset = 0
        shape_ind = None
        for shape_ind, shape_patch_count in enumerate(self.shape_patch_count):
            if index >= shape_patch_offset and index < shape_patch_offset + shape_patch_count:
                shape_patch_ind = index - shape_patch_offset
                break
            shape_patch_offset = shape_patch_offset + shape_patch_count

        return shape_ind, shape_patch_ind

    # load shape from a given shape index
    def load_shape_by_index(self, shape_ind):  
        prefix = os.path.join(self.root, self.shape_names[shape_ind])
        feat = np.load(prefix+'.m0.npy')            
        depth = np.load(prefix+'.depth.npy')
        return Shape(feat, depth)
        
       



