from __future__ import print_function
import argparse
import os
import sys
import random
import numpy as np
import torch
import torch.utils.data
from dataset import PatchDataset, DataSampler
from cnnnet import CNNnet


def parse_arguments():
    parser = argparse.ArgumentParser()

    # naming / file handling
    parser.add_argument('--indir', type=str, default='./mpm_feat', help='input folder (point clouds)')
    parser.add_argument('--outdir', type=str, default='./results/', help='output folder (estimated point cloud properties)')
    parser.add_argument('--dataset', type=str, default='testingset.txt', help='shape set file name')
    parser.add_argument('--modeldir', type=str, default='./models', help='model folder')
    parser.add_argument('--models', type=str, default='cnn_volume_singlescale', help='names of trained models, can evaluate multiple models')
    parser.add_argument('--modelpostfix', type=str, default='_model.pth', help='model file postfix')
    parser.add_argument('--parmpostfix', type=str, default='_params.pth', help='parameter file postfix')
    parser.add_argument('--gpu_idx', type=int, default=0, help='set < 0 to use CPU')

    
    parser.add_argument('--batchSize', type=int, default=128, help='batch size, if 0 the training batch size is used')
    parser.add_argument('--workers', type=int, default=4, help='number of data loading workers - 0 means same thread as main execution')
    parser.add_argument('--cache_capacity', type=int, default=100, help='Max. number of dataset elements (usually shapes) to hold in the cache at the same time.')
    parser.add_argument('--use_grad', action=argparse.BooleanOptionalAction, default=False, help='use gradient in loss function during training')
    parser.add_argument('--inputs', type=str, nargs='+', default=['.vol'], help='inputs of the network')
    parser.add_argument('--shape_names',type=str, nargs='+', default=[], help='name of a single shape')
    return parser.parse_args()

def eval_cnnnet(opt):

    opt.models = opt.models.split()

    device = torch.device("cpu" if opt.gpu_idx < 0 else "cuda:%d" % opt.gpu_idx)
    # print(torch.cuda.is_available())
    for model_name in opt.models:

        model_filename = os.path.join(opt.modeldir, model_name+opt.modelpostfix)
        param_filename = os.path.join(opt.modeldir, model_name+opt.parmpostfix)

        # load model and training parameters
        trainopt = torch.load(param_filename)

        if opt.batchSize == 0:
            model_batchSize = trainopt.batchSize
        else:
            model_batchSize = opt.batchSize

        # get indices in targets and predictions corresponding to each output
        pred_dim = 1
        dataset = PatchDataset(
            root=opt.indir, 
            shape_list_filename=opt.shape_names,
            cache_capacity=opt.cache_capacity,
            use_grad = False,
            use_depth = False,
            inputs = opt.inputs,
        )

        datasampler = DataSampler(dataset, shuffle_data=False)
        
        dataloader = torch.utils.data.DataLoader(
            dataset,
            sampler=datasampler,
            batch_size=model_batchSize,
            num_workers=int(opt.workers))

        in_dim = 0
        scale = 1
        for i in opt.inputs:
            if i == '.vol': in_dim += 1
            if i == '.cent': in_dim += 3
            if i == '.fine.vol': scale += 1
            if i == '.m0': in_dim += 1
            if i == '.m1': in_dim += 3
            if i == '.m2': in_dim += 6            

    
        regressor = CNNnet(
            input_dim=in_dim,
            scale=scale
        )

        regressor.load_state_dict(torch.load(model_filename))
        regressor.to(device)
        regressor.eval()

        shape_ind = 0
        shape_patch_offset = 0
        
        shape_patch_count = dataset.shape_patch_count[shape_ind]
        depth = torch.zeros(shape_patch_count, pred_dim, dtype=torch.float)

        # append model name to output directory and create directory if necessary
        model_outdir = os.path.join(opt.outdir, model_name)
        if not os.path.exists(model_outdir):
            os.makedirs(model_outdir)

        num_batch = len(dataloader)
        batch_enum = enumerate(dataloader, 0)
        for batchind, data in batch_enum:

            input = data[0]               
            # concatenate feats
            i = 0
            for s in range(scale):
                if(i == 0): input = data[0][i]
                else: input = torch.cat((input, data[0][i]), dim=1)              
                i += 1
                if(in_dim > 1): 
                    while(input.shape[1]<in_dim): 
                        input = torch.cat((input, data[0][i]), dim=1)                   
                        i += 1
                    
            pred = regressor(input)
            # print('[%s %d/%d] shape %s' % (model_name, batchind, num_batch-1, dataset.shape_names[shape_ind]))

            batch_offset = 0
            while batch_offset < pred.size(0):

                shape_patches_remaining = shape_patch_count-shape_patch_offset
                batch_patches_remaining = pred.size(0)-batch_offset

                # append estimated patch properties batch to properties for the current shape
                depth[shape_patch_offset:shape_patch_offset+min(shape_patches_remaining, batch_patches_remaining), :] = pred[
                    batch_offset:batch_offset+min(shape_patches_remaining, batch_patches_remaining), :].detach().cpu()
    

                batch_offset = batch_offset + min(shape_patches_remaining, batch_patches_remaining)
                shape_patch_offset = shape_patch_offset + min(shape_patches_remaining, batch_patches_remaining)

                if shape_patches_remaining <= batch_patches_remaining:                   
                    # save depth                  
                    np.savetxt(os.path.join(model_outdir, dataset.shape_names[shape_ind]+'.depth'), depth.numpy())
                                       
                    # start new shape
                    if shape_ind + 1 < len(dataset.shape_names):
                        shape_patch_offset = 0
                        shape_ind = shape_ind + 1                        
                        shape_patch_count = dataset.shape_patch_count[shape_ind]                      
                        depth = depth.new_zeros(shape_patch_count, pred_dim)

            

if __name__ == '__main__':
    eval_opt = parse_arguments()
    eval_cnnnet(eval_opt)
