#include <Core/Grids/GRID.h>
#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Spatial_Acceleration/Nearest_Neighbors.h>

#ifdef USE_TORCH
#include <torch/script.h>
#include <torch/torch.h>
#endif

using namespace PhysBAM;
typedef double T;
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;
typedef VECTOR<T,6> SV;

#ifdef USE_TORCH
int prec=6;

torch::Device device(torch::kCPU);
torch::Tensor m0_t;

T Poly6Kernel(T h,T x)
{
    if(x>=h) return 0;
    T c=(T)315/(64*pi*pow<9>(h));  // not use pow
    T hx=h*h-x*x,hx2=c*hx*hx;
    return hx*hx2;
}

void Compute_Density(const ARRAY<TV>& X,ARRAY<T> &densities,T m0_r)
{
    Nearest_Neighbors<TV,T>(X,
        [m0_r,&densities](int i,int j,T d)
        {
            T d_v=Poly6Kernel(m0_r,d);
            densities(i)+=d_v;
            densities(j)+=d_v;
        },m0_r);
}

void Find_Nodes_Inside_Surface(const GRID<TV>& grid,const ARRAY<TV>& X,int acc_radius,int threshold,ARRAY<bool,IV> &i_flags,ARRAY<bool,IV> &o_flags)
{
    ARRAY<int,IV> p_flags(grid.Cell_Indices(acc_radius*2));
    for(int i=0;i<X.m;i++)
    {
        auto cell=grid.Cell_I(X(i));
        if(p_flags.Valid_Index(cell))
            p_flags(cell)=1;
    }

    ARRAY<int,IV> table(grid.Node_Indices(acc_radius*2));
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices(acc_radius));it.Valid();it.Next())
    {
        int x=it.index.x;int y=it.index.y; int z=it.index.z;
        table(it.index)=table(IV(x-1,y,z))+table(IV(x,y-1,z))+table(IV(x,y,z-1))
            -table(IV(x-1,y-1,z))-table(IV(x,y-1,z-1))-table(IV(x-1,y,z-1))
            +table(it.index-1)+p_flags(it.index-1);
    }

    int sum_i=0,sum_o=0;
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices());it.Valid();it.Next())
    {
        IV pos=it.index+acc_radius;
        int x=pos.x;int y=pos.y;int z=pos.z;
        int diameter=2*acc_radius;
        // determine inside nodes
        int neighbors=table(pos)-table(IV(x-diameter,y,z))
            -table(IV(x,y-diameter,z))-table(IV(x,y,z-diameter))
            +table(IV(x-diameter,y-diameter,z))+table(IV(x,y-diameter,z-diameter))
            +table(IV(x-diameter,y,z-diameter))-table(pos-diameter);

        if(neighbors>threshold)
        {
            i_flags(it.index)=true;
            sum_i++;
            continue;
        }
        //determine outside nodes
        pos=it.index+acc_radius/2;
        x=pos.x;y=pos.y;z=pos.z;
        neighbors=table(pos)-table(IV(x-acc_radius,y,z))
            -table(IV(x,y-acc_radius,z))-table(IV(x,y,z-acc_radius))
            +table(IV(x-acc_radius,y-acc_radius,z))+table(IV(x,y-acc_radius,z-acc_radius))
            +table(IV(x-acc_radius,y,z-acc_radius))-table(pos-acc_radius);
        if(neighbors==0)
        {
            o_flags(it.index)=true;
            sum_o++;
        }
    }
}

void Compute_Batch(torch::jit::script::Module& cnnnet,torch::Tensor& x,
    const ARRAY<IV>& nn_indices,ARRAY<T,IV>& node_phi,IV offset,int stride,
    T dx,int batch_count)
{
    std::vector<torch::jit::IValue> x_std;
    x_std.push_back(x);

    torch::Tensor depth=cnnnet.forward(x_std).toTensor().to(torch::kFloat64);
    torch::Tensor depth_cpu = depth.to(device);
    double* ptr=(double*)depth_cpu.data_ptr();
    int len=depth_cpu.numel();
    PHYSBAM_ASSERT(len>=batch_count);
    for(int i=0;i<batch_count;i++)
        node_phi(nn_indices(i)*stride+offset)=ptr[i]*dx;
}

void Compute_m0(ARRAY<T,IV>& m0,const GRID<TV>& subgrid,const ARRAY<TV>& X,
    const ARRAY<T>& densities,int m0_radius)
{
    T dx=subgrid.dX(0),m0_r=dx*m0_radius;
    for(int i=0;i<X.m;i++)
    {
        RANGE<IV> neighbors(subgrid.Cell_I(X(i)));
        neighbors=neighbors.Thickened(m0_radius).To_Half_Opened();
        neighbors=m0.domain.Intersect(neighbors);
        for(RANGE_ITERATOR<TV::m> it(neighbors);it.Valid();it.Next())
        {
            T dist=(subgrid.Cell_X(it.index)-X(i)).Magnitude();
            T m0_v=Poly6Kernel(m0_r,dist);
            m0(it.index)+=m0_v/densities(i);
        }
    }
}

void Construct_Features_Subgrid(torch::jit::script::Module& cnnnet,
    ARRAY<T,IV>& node_phi,IV offset,int stride,
    const GRID<TV>& subgrid,const ARRAY<TV>& X,
    int batch_size,const ARRAY<T>& densities,int m0_radius,
    int acc_radius,int fw,int threshold)
{
    T dx=subgrid.dX(0);

    // acceleration
    ARRAY<bool,IV> i_flags(subgrid.Node_Indices());
    ARRAY<bool,IV> o_flags(subgrid.Node_Indices());
    Find_Nodes_Inside_Surface(subgrid,X,acc_radius,threshold,i_flags,o_flags);

    // compute m0 features
    ARRAY<T,IV> m0(subgrid.Cell_Indices(2*fw));
    Compute_m0(m0,subgrid,X,densities,m0_radius);

    torch::Tensor x=at::empty({batch_size,1,2*fw,2*fw,2*fw},torch::kFloat64);
    ARRAY<IV> nn_indices(batch_size);

    // save all node indices to an array
    ARRAY<IV> node_indices;
    for(RANGE_ITERATOR<TV::m> ir(subgrid.Node_Indices());ir.Valid();ir.Next())
    {
        if(o_flags(ir.index))
            node_phi(ir.index*stride+offset)=2*dx;
        else if(i_flags(ir.index))
            node_phi(ir.index*stride+offset)=-2*dx;
        else node_indices.Append(ir.index);
    }

    int batch_count=0;
    auto options = torch::TensorOptions().dtype(torch::kFloat64).device(torch::kCPU);
    for(auto p:node_indices)
    {
        m0_t = torch::from_blob(m0.Get_Array_Pointer(), {m0.Size()(0),m0.Size()(1),m0.Size()(2)}, options);

        IV i=p+2*fw;
        nn_indices(batch_count)=p;

        x[batch_count][0]=m0_t.index({torch::indexing::Slice(i.x-fw,i.x+fw),torch::indexing::Slice(i.y-fw,i.y+fw),torch::indexing::Slice(i.z-fw,i.z+fw)});

        if(++batch_count==batch_size)
        {
            Compute_Batch(cnnnet,x,nn_indices,node_phi,offset,stride,dx,batch_count);
            batch_count=0;
        }
    }
    if(batch_count>0)
        Compute_Batch(cnnnet,x,nn_indices,node_phi,offset,stride,dx,batch_count);
}

void Construct_Features(ARRAY<T,IV>& node_phi,const GRID<TV>& grid,
    const ARRAY<TV>& X,const std::string& file_model,
    int batch_size,const ARRAY<T>& densities,int stride,int m0_radius,
    int acc_radius,int feat_radius,int threshold)
{
    T dx=grid.dX.Min();
    torch::jit::script::Module cnnnet=torch::jit::load(file_model);

    // loop over subgrids
    for(RANGE_ITERATOR<TV::m> it(IV()+stride);it.Valid();it.Next())
    {
        IV counts=(grid.numbers_of_cells-it.index)/stride;
        TV L=grid.domain.min_corner+TV(it.index)*dx,U=L+stride*dx*TV(counts);
        GRID<TV> subgrid(counts,{L,U});

        Construct_Features_Subgrid(cnnnet,node_phi,it.index,stride,subgrid,X,
            batch_size,densities,m0_radius,acc_radius,feat_radius,threshold);
    }
}

int main(int argc,char *argv[])
{
    std::string file_xyz,file_model,file_phi,file_grid;
    int threshold=410; // node is inside if cells with particles>threshold
    int batch_size=256;
    int acc_radius=4; // acceleration radius
    int m0_radius=3;
    int feat_radius=4;
    int stride=2;
    T dx=0;
    TV lb;
    IV N;

    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-dx",&dx,"dx","phi grid dx");
    parse_args.Add("-lb",&lb,"x y z","location of lower corner");
    parse_args.Add("-N",&N,"m n p","numbers of grid cells");
    parse_args.Add("-stride",&stride,"stride","stride of phi");
    parse_args.Add("-i",&file_xyz,"file","Input xyz positions file");
    parse_args.Add("-model",&file_model,"file","Input neural network model");
    parse_args.Add("-o",&file_phi,"file","Output file for phi");
    parse_args.Add("-grid",&file_grid,"file","Output file for grid");

    parse_args.Add("-feat_radius",&feat_radius,"width","Radius of cells for network");
    parse_args.Add("-acc_radius",&acc_radius,"radius","Radius of acceleration");
    parse_args.Add("-threshold",&threshold,"threshold","Thresholdx of acceleration");
    parse_args.Add("-m0_r",&m0_radius,"radius","Radius of m0 feature in grid cells");
    parse_args.Add("-batch_size",&batch_size,"batch size","batch size of the network");
    parse_args.Parse();

    PHYSBAM_ASSERT(dx>0);
    PHYSBAM_ASSERT(N.Min()>=2*stride);    
    PHYSBAM_ASSERT(file_xyz.length()>0);    
    PHYSBAM_ASSERT(file_model.length()>0);    
    PHYSBAM_ASSERT(file_phi.length()>0);    
    PHYSBAM_ASSERT(file_grid.length()>0);
    
    ARRAY<TV> X;
    Read_From_File(file_xyz,X);

    GRID<TV> grid(N,RANGE<TV>(lb,lb+dx*TV(N)));
    T m0_r=stride*dx*m0_radius;
    T d_z=Poly6Kernel(m0_r,0);
    ARRAY<T> densities(X.m,init_new,d_z);
    Compute_Density(X,densities,m0_r);

    ARRAY<T,IV> node_phi(grid.Node_Indices());
    Construct_Features(node_phi,grid,X,file_model,batch_size,densities,stride,m0_radius,acc_radius,feat_radius,threshold);

    Write_To_File(Stream_Type(0.),file_phi,node_phi);
    Write_To_File(Stream_Type(0.),file_grid,grid);
    LOG::cout<<"Constructing surface from particles in "<<file_xyz<<std::endl;
}
#else
int main(int argc,char *argv[])
{
    PHYSBAM_FATAL_ERROR("You must enable USE_TORCH to use this.");
    return 0;
}
#endif
