#include <Core/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Meshes/SEGMENTED_CURVE.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>

using namespace PhysBAM;
using std::sin;
using std::cos;
typedef double T;
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;
typedef VECTOR<T,2> TV2;

int main(int argc,char *argv[])
{
    T x0=.2;
    T x1=.01;
    int frame=100;
    std::string output="output";
    
    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-x0",&x0,"dx","initial particle separation");
    parse_args.Add("-x1",&x1,"dx","final particle separation");
    parse_args.Add("-frame",&frame,"frames","number of frames to run.");
    parse_args.Add("-o",&output,"dir","output directory");
    parse_args.Parse();

    // // Set opengl viewer
    VIEWER_DIR viewer_dir(output);
    VIEWER_OUTPUT viewer_output(Stream_Type(0.),viewer_dir);
    viewer_output.template Add_Axes<TV::m>("axes",{});

    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(1234);
    
    SEGMENTED_CURVE<TV> box;
    box.particles.Add_Elements(8);
    int k=0;
    for(RANGE_ITERATOR<TV::m> it(IV()+2);it.Valid();it.Next())
        box.particles.X(k++)=TV(it.index);
    for(int a=0;a<TV::m;a++)
        for(int i=0;i<(1<<TV::m);i++)
            if(!(i&(1<<a)))
                box.mesh.elements.Append({i,i+(1<<a)});
    box.Update_Number_Nodes();
    viewer_output.Add("box",box,true,{.draw=true,.color={{1,1,0}}});

    T e=(T).5/x1;
    
    ARRAY<TV2> X2;
    POISSON_DISK<TV2> poisson_disk(1);
    poisson_disk.Sample(rand,RANGE<TV2>::Centered_Box()*e,X2);
    LOG::printf("particles: %P\n",X2.m);
    ARRAY<TV> X0(X2.m);
    for(int i=0;i<X2.m;i++) X0(i)=X2(i).Append(0);

    GEOMETRY_PARTICLES<TV> particles;    
    particles.Add_Elements(X0.m);

    viewer_output.Add("particles",particles,false,{.draw=true});
    for(int f=0;f<=frame;f++)
    {
        T w=x0*pow(x1/x0,(T)f/frame);
        ARRAY<TV> X1;
        X1=X0*w+(TV()+(T).5);
        for(int i=X1.m-1;i>=0;i--)
            if(RANGE<TV>::Unit_Box().Lazy_Outside(X1(i)))
            {
                X1.Remove_Index_Lazy(i);
                continue;
            }
        
        particles.Resize(X1.m);
        particles.X=X1;
        Flush_Frame(LOG::sprintf("%d",f).c_str());
        Write_To_File(Stream_Type(0.),LOG::sprintf("%s/X-%03d.xyz",output,f),particles.X);
    }
    return 0;
}
