#include <Core/Analytic_Tests/ANALYTIC_SCALAR.h>
#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <Core/Grids/GRID.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Log/LOG.h>
#include <Core/Math_Tools/cube.h>
#include <Core/Nonlinear_Equations/ITERATIVE_SOLVER.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include <cmath>
#include <fstream>
#include <functional>

using namespace PhysBAM;
using std::sin;
using std::cos;
typedef double T;
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;
T period=3;

TV v(TV point,T t)
{
    T x=point(0);T y=point(1);T z=point(2);
    TV velosity;
    velosity(0)=2*sqr(sin(pi*x))*sin(2*pi*y)*sin(2*pi*z)*cos(pi*t/period);
    velosity(1)=-sin(2*pi*x)*sqr(sin(pi*y))*sin(2*pi*z)*cos(pi*t/period);
    velosity(2)=-sin(2*pi*x)*sin(2*pi*y)*sqr(sin(pi*z))*cos(pi*t/period);
    return velosity;
}

void Update_V(const ARRAY<TV>& X,ARRAY<TV>& V,T t)
{
    for(int i=0;i<X.m;i++)
    {
        TV point=X(i);
        V(i)=v(point,t);
    }    
}

void Euler_Step(ARRAY<TV>& X,ARRAY<TV>& V,T dt,T t,int rk_order)
{
    for(int i=0;i<X.m;i++)
    {
        if(rk_order==1)X(i)+=dt*V(i);
        if(rk_order==3)
        {
            TV k1=v(X(i),t);
            TV k2=v(X(i)+dt*k1/2,t+dt/2);
            TV k3=v(X(i)-dt*k1+2*dt*k2,t+dt);
            TV k=(k1+4*k2+k3)/6;
            X(i)+=dt*k;
        }
    }    
}

void Dump_Particles(ARRAY<TV>& X)
{
    for(int i=0;i<X.m;i++) Add_Point(X(i),{1,0,0});
}

ANALYTIC_SCALAR<TV> * as = nullptr;

ANALYTIC_SCALAR<TV>* as_list[]=
{
    Make_Analytic_Scalar<TV>([](auto X,auto t){return (X-(TV()+(T)0.35)).Magnitude()-0.15;}),
    Make_Analytic_Scalar<TV>([](auto X,auto t){return sqr(X-(TV()+(T)0.35)).Magnitude_Squared()-sqr(sqr(0.15));}),
    Make_Analytic_Scalar<TV>([](auto X,auto t){
        auto x=(X.Dot(TV(1,0,0))-(T).35)/(T).15;
        auto y=(X.Dot(TV(0,1,0))-(T).35)/(T).15;
        auto z=(X.Dot(TV(0,0,1))-(T).35)/(T).15;
        T a=-.2;
        return 60*y*(a*a+a+1)*(a*a+3*a+3)*z*z*z*(x*x-1./3*y*y)*(a+2)*a*sqrt(2)+(y*y*y*y*y*y+(39*x*x-21*z*z)*y*y*y*y+(-21*x*x*x*x-42*x*x*z*z+24*z*z*z*z)*y*y+5*x*x*x*x*x*x-21*x*x*x*x*z*z+24*x*x*z*z*z*z-5)*a*a*a*a*a*a+(6*y*y*y*y*y*y+(234*x*x-126*z*z)*y*y*y*y+(-126*x*x*x*x-252*x*x*z*z+144*z*z*z*z)*y*y+30*x*x*x*x*x*x-126*x*x*x*x*z*z+144*x*x*z*z*z*z-30)*a*a*a*a*a+(15*y*y*y*y*y*y+(585*x*x-315*z*z)*y*y*y*y+(-315*x*x*x*x-630*x*x*z*z+360*z*z*z*z)*y*y+75*x*x*x*x*x*x-315*x*x*x*x*z*z+360*x*x*z*z*z*z-75)*a*a*a*a+(20*y*y*y*y*y*y+(780*x*x-420*z*z)*y*y*y*y+(-420*x*x*x*x-840*x*x*z*z+480*z*z*z*z)*y*y+100*x*x*x*x*x*x-420*x*x*x*x*z*z+480*x*x*z*z*z*z-100)*a*a*a+(15*y*y*y*y*y*y+(585*x*x-315*z*z)*y*y*y*y+(-315*x*x*x*x-630*x*x*z*z+360*z*z*z*z)*y*y+75*x*x*x*x*x*x-315*x*x*x*x*z*z+360*x*x*z*z*z*z-75)*a*a+(6*y*y*y*y*y*y+(234*x*x-126*z*z)*y*y*y*y+(-126*x*x*x*x-252*x*x*z*z+144*z*z*z*z)*y*y+30*x*x*x*x*x*x-126*x*x*x*x*z*z+144*x*x*z*z*z*z-30)*a+5*(y*y*y*y+(2*x*x+2*z*z+1)*y*y+x*x*x*x+(2*z*z+1)*x*x+z*z*z*z+z*z+1)*(x*x+y*y+z*z-1);}),
};

RANGE<TV> seed_box_list[]=
{
    {TV()+(T)0.35-(T).15,TV()+(T)0.35+(T).15},
    {TV()+(T)0.35-(T).15,TV()+(T)0.35+(T).15},
    {TV()+(T)0.35-(T).15,TV()+(T)0.35+(T).15},
};

int main(int argc,char *argv[])
{
    int resolution=100;
    const unsigned int seed=123;
    int rk_order=3; //runge-kutta order
    T ppc(8);
    std::string output;
    int starting_levelset=0;
    int frame=1500;
    
    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-r",&resolution,"res","resolution");
    parse_args.Add("-o",&output,"output dir","Output directory of xyz and viewer");
    parse_args.Add("-rk_order",&rk_order,"rk_order","runge kutta order");
    parse_args.Add("-ls",&starting_levelset,"index","index of initial level set.");
    parse_args.Add("-frame",&frame,"frames","number of frames to run.");
    parse_args.Parse();
    GRID<TV> grid(IV()+resolution,RANGE<TV>::Unit_Box());

    as = as_list[starting_levelset];
    ARRAY<TV> X;
    
    // Generate points on sphere surface
    RANDOM_NUMBERS<T> rand(seed);
    {
        POISSON_DISK<TV> poisson_disk(1);
        poisson_disk.Set_Distance_By_Volume(grid.dX.Product()/ppc);
        poisson_disk.Sample(rand,seed_box_list[starting_levelset],X);
        for(int i=X.m-1;i>=0;i--)
            if(as->f(X(i),0)>0)
                X.Remove_Index_Lazy(i);
        
        LOG::printf("particles: %P\n",X.m);
    }
    

    ARRAY<TV> V(X.m);

    // // Set opengl viewer
    VIEWER_DIR viewer_dir(output);
    VIEWER_OUTPUT viewer_output(Stream_Type(0.),viewer_dir);
    viewer_output.template Add_Axes<TV::m>("axes",{});
    viewer_output.Add("debug_particles",DEBUG_PARTICLES<TV>::Get(),false,{.draw=false});
    viewer_output.Add("grid",grid,true,{.draw=false});
    
    // // Forward Euler     
    T dt=period/(2*frame);
    int count=0;

    fs::create_directories(output);    
    Dump_Particles(X);
    Flush_Frame(LOG::sprintf("%d",0).c_str());
    Write_To_File(Stream_Type(0.),LOG::sprintf("%s/X-%03d.xyz",output,count++),X);
        
    for(int i=0;i<frame;i++)
    {
        T t=i*dt;
        Update_V(X,V,t);
        Euler_Step(X,V,dt,t,rk_order);

        if(i%10==0)
        {
            Dump_Particles(X);
            Flush_Frame(LOG::sprintf("%d",i/100+1).c_str());
            Write_To_File(Stream_Type(0.),LOG::sprintf("%s/X-%03d.xyz",output,count++),X);
        }
    }
}
