#include <Core/Basic_Geometry/BOWL.h>
#include <Core/Basic_Geometry/CONE.h>
#include <Core/Basic_Geometry/CYLINDER.h>
#include <Core/Basic_Geometry/ELLIPSOID.h>
#include <Core/Basic_Geometry/ORIENTED_BOX.h>
#include <Core/Basic_Geometry/PLANE.h>
#include <Core/Basic_Geometry/RING.h>
#include <Core/Basic_Geometry/SEGMENT_3D.h>
#include <Core/Basic_Geometry/SPHERE.h>
#include <Core/Basic_Geometry/TORUS.h>
#include <Core/Basic_Geometry/TRIANGLE_3D.h>
#include <Core/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Core/Grids/GRID.h>
#include <Core/Level_Sets/LEVELSET.h>
#include <Core/Parsing/PARSE_ARGS.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Seeding/POISSON_DISK.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Viewer/VIEWER_OUTPUT.h>
#include <fstream>
#include <functional>

using namespace PhysBAM;
typedef double T;
typedef VECTOR<T,3> TV;
typedef VECTOR<int,3> IV;

T thin_thickness;
T small_radius_hi, small_radius_lo;
T small_height_hi, small_height_lo;
T inner_radius_lo, inner_radius_hi;
T outer_radius;
T dx;
bool thin_surface=0;
T thickness=0;
T gap=0.5;

template<class TV>
struct OBJECT
{
    typedef typename TV::SCALAR T;

    RANGE<TV> box;

    virtual ~OBJECT()=default;
    virtual T phi(const TV& X) const=0;
    virtual TV grad_phi(const TV& X) const=0;
};

template<class OBJ>
struct OBJECT_SD:OBJECT<TV>
{
    OBJ o;
    OBJECT_SD(RANDOM_NUMBERS<T>& r,const RANGE<TV>& b)
    {
        do{
            r.Fill_Uniform(o,b);
        }while(o.Volume()<b.Size()/128);
         box=o.Bounding_Box();
         
    }
    T phi(const TV& X) const override {return o.Signed_Distance(X);}
    TV grad_phi(const TV& X) const override {return o.Normal(X);}
};

template<class OBJ>
struct OBJECT_SD2:OBJECT<TV>
{
    OBJ o;
    OBJECT_SD2(RANDOM_NUMBERS<T>& r,const RANGE<TV>& b)
    {
        do{
            r.Fill_Uniform(o,b);
        }while(o.Volume()<b.Size()/64);
         box=o.Bounding_Box();         
    }
    T phi(const TV& X) const override {return o.Approximate_Signed_Distance(X);}
    TV grad_phi(const TV& X) const override {return o.Normal(X);}
};


template<class OBJ>
struct OBJECT_DTE:OBJECT<TV>
{
    OBJ o;
    OBJECT_DTE(RANDOM_NUMBERS<T>& r,const RANGE<TV>& b)
    {
        r.Fill_Uniform(o,b);
        box=o.Bounding_Box();
        box.Change_Size(thin_thickness);
    }
    T phi(const TV& X) const override
    {return o.Distance_To_Element(X)-thin_thickness;}
    TV grad_phi(const TV& X) const override
    {return o.Diff_Distance_To_Element(X).second;}
};



struct SMALL_SPHERE:OBJECT<TV>
{
    SPHERE<TV> o;
    SMALL_SPHERE(RANDOM_NUMBERS<T>& r,const RANGE<TV>& b)
    {
        r.Fill_Uniform(o.center,b.Thickened(-small_radius_hi));
        r.Fill_Uniform(o.radius,small_radius_lo,small_radius_hi);
        box=o.Bounding_Box();
    }
    T phi(const TV& X) const override {return o.Signed_Distance(X)-dx;}
    TV grad_phi(const TV& X) const override {return o.Normal(X);}
};

struct SMALL_RING:OBJECT<TV>
{
    RING<T> o;
    SMALL_RING(RANDOM_NUMBERS<T>& r,const RANGE<TV>& b)
    {
        o.plane1.normal=r.template Get_Direction<TV>();
        o.plane1.normal.Normalize();
        o.plane2.normal=-o.plane1.normal;
        r.Fill_Uniform(o.height,small_height_lo,small_height_hi);
        o.Set_Height(o.height);
        r.Fill_Uniform(o.inner_radius,inner_radius_lo,inner_radius_hi);
        o.outer_radius=outer_radius;
        box=o.Bounding_Box();
    }
    T phi(const TV& X) const override {return o.Signed_Distance(X);}
    TV grad_phi(const TV& X) const override {return o.Normal(X);}
};


ARRAY<std::pair<std::string,std::function<OBJECT<TV>*(RANDOM_NUMBERS<T>& r,const RANGE<TV>& box)>>> object_type =
{
    {"sphere",[](auto&r,auto&b){return new OBJECT_SD<SPHERE<TV>>(r,b);}},
    {"cone",[](auto&r,auto&b){return new OBJECT_SD<CONE<T>>(r,b);}},
    {"torus",[](auto&r,auto&b){return new OBJECT_SD<TORUS<T>>(r,b);}},
    {"triangle",[](auto&r,auto&b){return new OBJECT_DTE<TRIANGLE_3D<T>>(r,b);}},
    {"segment",[](auto&r,auto&b){return new OBJECT_DTE<SEGMENT_3D<T>>(r,b);}},
    {"smaller_sphere",[](auto&r,auto&b){return new SMALL_SPHERE(r,b);}},
    {"smaller_ring",[](auto&r,auto&b){return new SMALL_RING(r,b);}},    
    {"ring",[](auto&r,auto&b){return new OBJECT_SD<RING<T>>(r,b);}},
    {"cylinder",[](auto&r,auto&b){return new OBJECT_SD<CYLINDER<T>>(r,b);}},
    {"oriented_box",[](auto&r,auto&b){return new OBJECT_SD<ORIENTED_BOX<TV>>(r,b);}},
    {"ellipsoid",[](auto&r,auto&b){return new OBJECT_SD2<ELLIPSOID<TV>>(r,b);}}
};

template<class T,class TV,class IV>
void Sample_Object_Levelset(const GRID<TV>& grid,const OBJECT<TV>* o,ARRAY<T,IV>& node_phi,ARRAY<TV,IV>& grad_node_phi,bool use_grad,bool inv_sample)
{
    auto seed_box=grid.domain.Intersect(o->box.Thickened(grid.dX.Min()));
    ORIENTED_BOX<TV> box(seed_box,ROTATION<TV>());
    for(RANGE_ITERATOR<TV::m> it(node_phi.domain);it.Valid();it.Next())
    {
        if(thin_surface)
        {node_phi(it.index)=abs(o->phi(grid.Node_X(it.index)))-thickness-gap*grid.dX.Min();}
        else
        {
            TV node_x=grid.Node_X(it.index);
            node_phi(it.index)=inv_sample?max(box.Signed_Distance(node_x),-(o->phi(node_x))):o->phi(node_x);
        }
        if(use_grad)
          grad_node_phi(it.index)=o->grad_phi(grid.Node_X(it.index));
    }
}

template<class T,class TV>
void Sample_Points_In_Object(const GRID<TV>& grid,const OBJECT<TV>* o,
    RANDOM_NUMBERS<T>& rand,POISSON_DISK<TV>& poisson_disk,ARRAY<TV>& X,bool enable_jitter,bool inv_sample)
{

    auto seed_box=grid.domain.Intersect(o->box.Thickened(grid.dX.Min()));    
    ORIENTED_BOX<TV> box(seed_box,ROTATION<TV>());
    poisson_disk.Sample(rand,seed_box,X);
    if(X.m<=1) X={o->box.Center()};

    if(!thin_surface) thickness-=gap*grid.dX.Min();
    for(int i=X.m-1;i>=0;i--)
    {
        if(enable_jitter)
            X(i)+=min(2*grid.dX.Min(),abs(o->phi(X(i))))*rand.template Get_Direction<TV>();
        T phi=inv_sample?max(box.Signed_Distance(X(i)),-o->phi(X(i))):o->phi(X(i));
        if(phi>thickness ||(thin_surface && phi<-thickness))
        {
            X.Remove_Index_Lazy(i);
            continue;
        }
    }
    
}

int main(int argc,char *argv[])
{
    int seed=123;
    int resolution=50;
    T cpp=5000,delta=0;
    T particles_per_cell=1<<TV::m;
    std::string shape;
    std::string file_xyz,file_grid,file_node_phi,file_grad_node_phi;
    std::string prefix;
    bool use_grad=0,enable_jitter=0,use_ip=0,inv_sample=0;

    PARSE_ARGS parse_args(argc,argv);
    parse_args.Add("-seed",&seed,"seed","Random number seed");
    parse_args.Add("-resolution",&resolution,"res","Grid resolution.");
    parse_args.Add("-shape",&shape,"shape","Shape used for training set");
    parse_args.Add("-ppc",&particles_per_cell,"ppc","Particles per cell");
    parse_args.Add("-cpp",&cpp,"cpp","Cells per particle");
    parse_args.Add("-o_prefix",&prefix,"file prefix","Prefix of outputfiles");
    parse_args.Add("-use_grad",&use_grad,"generate grad phi");
    parse_args.Add("-use_ip",&use_ip,"generate with isolated particles");
    parse_args.Add("-thin_surface",&thin_surface,"generate only a thin surface of object");
    parse_args.Add("-delta",&delta,"thickness","the thickness of thin surface");
    parse_args.Add("-gap",&gap,"gap","the gap between the surface and the particles");
    parse_args.Add("-enable_jitter",&enable_jitter,"enable jittering in training data");
    parse_args.Add("-inv_sample",&inv_sample,"enable jittering in training data");
    parse_args.Parse();
    
    file_xyz=prefix+".xyz";
    file_grid=prefix+".grid";
    file_node_phi=prefix+".phi";
    file_grad_node_phi=prefix+".grad.phi";

    RANDOM_NUMBERS<T> rand;
    rand.Set_Seed(seed);
    GRID<TV> grid(IV()+resolution,RANGE<TV>::Unit_Box());
    dx=grid.dX.Min();
    
    if(thin_surface) thickness=delta*dx;
    thin_thickness=dx*0.5;
    small_radius_lo=dx*(T).5; small_radius_hi=dx*2;
    small_height_lo=2*dx; small_height_hi=4*dx;
    inner_radius_lo=dx; inner_radius_hi=4*dx;
    outer_radius=10*dx;

    
    POISSON_DISK<TV> poisson_disk(1);
    poisson_disk.Set_Distance_By_Volume(grid.dX.Product()/particles_per_cell);

    int shape_id=-1;
    for(int i=0;i<object_type.m;i++)
        if(object_type(i).first==shape)
            shape_id=i;
    PHYSBAM_ASSERT(shape_id>=0);

    RANGE<TV> sample_range;
    
    sample_range=grid.domain.Thickened(-grid.dX.Min());
    OBJECT<TV> * obj=object_type(shape_id).second(rand,sample_range);

    ARRAY<TV> X;
    Sample_Points_In_Object(grid,obj,rand,poisson_disk,X,enable_jitter,inv_sample);

    ARRAY<T,IV> node_phi(grid.Node_Indices().To_Half_Opened());
    ARRAY<TV,IV> grad_node_phi;
    if(use_grad) grad_node_phi.Resize(grid.Node_Indices());
    Sample_Object_Levelset(grid,obj,node_phi,grad_node_phi,use_grad,inv_sample);

    if(use_ip)
    {
        RANGE<TV> seed_box_ip=grid.domain.Intersect(obj->box.Thickened(4*dx));
        POISSON_DISK<TV> poisson_disk_ip(1);
        poisson_disk_ip.Set_Distance_By_Volume(grid.dX.Product()*cpp);

        ARRAY<TV> X_ip;
        poisson_disk_ip.Sample(rand,seed_box_ip,X_ip);

        
        for(int i=X_ip.m-1;i>=0;i--)
        {
            T phi_ip=obj->phi(X_ip(i));
            if(phi_ip<3*dx)
            {                
                X_ip.Remove_Index_Lazy(i);
                continue;
            }            
        }

        for(int i=X_ip.m-1;i>=0;i--)
        {
            
            SPHERE<TV> o(X_ip(i),1.1*dx);
            IV index=grid.Node_I(X_ip(i));
            RANGE<IV> grid_ip(index-2,index+3);
            grid_ip=grid.Node_Indices().Intersect(grid_ip);
            for(RANGE_ITERATOR<IV::m> ir(grid_ip);ir.Valid();ir.Next())
            {
                TV pos=grid.Node_X(ir.index);
                node_phi(ir.index)=min(node_phi(ir.index),o.Signed_Distance(pos));
            }            
        }
        
        X.Resize(X.m+X_ip.m);
        X.Append(X_ip);        
    }
    
    delete obj;
    
    Write_To_File(Stream_Type(0.),file_xyz,X);
    Write_To_File(Stream_Type(0.),file_grid,grid);
    Write_To_File(Stream_Type(0.),file_node_phi,node_phi);
    if(use_grad)
        Write_To_File(Stream_Type(0.),file_grad_node_phi,grad_node_phi);

        
    return 0;
}
