#!/usr/bin/env python3

import argparse
import os
import random
import math
import torch
import numpy as np
import time
from tensorboardX import SummaryWriter
from dataset import DataSampler
from dataset import PatchDataset
from cnnnet import CNNnet
from pathlib import Path
from lsr_mat import lsr_matrix

torch.autograd.set_detect_anomaly(False)
lsr_matrix = lsr_matrix.to(torch.float64)

def parse_arguments():
    parser = argparse.ArgumentParser()
    # naming / file handling
    parser.add_argument('--name', type=str,
                        default='test_script',help='trained model name')
    parser.add_argument('--indir', type=str,
                        default=os.path.dirname(os.path.abspath(__file__))+'/cnn_dataset',
                        help='input folder (point clouds)')
    parser.add_argument('--outdir', type=str,
                        default='./models',help='output folder (trained models)')
    parser.add_argument('--logdir', type=str,
                        default='./logs', help='training log folder')
    parser.add_argument('--trainset',type=str,
                        nargs='+', default=["training.0"], help='names of training data')
    parser.add_argument('--testset',type=str,
                        nargs='+', default=["validation.0"], help='names of testing data')
    parser.add_argument('--saveinterval', type=int,
                        default='10', help='save model each n epochs')
    parser.add_argument('--width', type=int,
                        default='8', help='feature width')
    parser.add_argument('--gpu_idx', type=int, default=0,
                        help='set < 0 to use CPU')

    # training parameters
    parser.add_argument('--nepoch', type=int,
                        default=5, help='number of epochs to train for')
    parser.add_argument('--batchSize', type=int,
                        default=256, help='input batch size')
    parser.add_argument('--workers', type=int,
                        default=16,
                        help='number of data loading workers - 0 means same thread as main execution')
    parser.add_argument('--cache_capacity', type=int,
                        default=150,
                        help='Max number of dataset elements in cache')
    parser.add_argument('--seed', type=int,
                        default=3627473, help='manual seed')
    parser.add_argument('--lr', type=float,
                        default=0.0002, help='learning rate')
    parser.add_argument('--momentum', type=float, default=0.9,
                        help='gradient descent momentum')
    return parser.parse_args()

def train_cnnnet(opt):
    device = torch.device("cpu" if opt.gpu_idx < 0
                          else "cuda:%d" % opt.gpu_idx)

    # colored console output
    def green(x): return '\033[92m' + x + '\033[0m'
    def blue(x): return '\033[94m' + x + '\033[0m'

    log_dirname = os.path.join(opt.logdir, opt.name)
    model_filename = os.path.join(opt.outdir, '%s.pth' % (opt.name))

    # create model
    w = opt.width    
    cnnnet = CNNnet(
        length=opt.width)

    if opt.seed < 0:
        opt.seed = random.randint(1, 10000)

    print("Random Seed: %d" % (opt.seed))
    random.seed(opt.seed)
    torch.manual_seed(opt.seed)

    # create train and test dataset loaders
    train_dataset = PatchDataset(
        root=opt.indir,
        shape_names=opt.trainset,
        cache_capacity=opt.cache_capacity
    )

    train_datasampler = DataSampler(
        train_dataset,
        shuffle_data=True)

    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        sampler=train_datasampler,
        batch_size=opt.batchSize,
        num_workers=int(opt.workers),
        drop_last=True)

    test_dataset = PatchDataset(
        root=opt.indir,
        shape_names=opt.testset,
        cache_capacity=opt.cache_capacity
        )

    test_datasampler = DataSampler(
        test_dataset,
        shuffle_data=True)

    test_dataloader = torch.utils.data.DataLoader(
        test_dataset,
        sampler=test_datasampler,
        batch_size=opt.batchSize,
        num_workers=int(opt.workers),
        drop_last=True)

    print('training set: %d patches (in %d batches) - test set: %d patches (in %d batches)' %
          (len(train_datasampler), len(train_dataloader), len(test_datasampler), len(test_dataloader)))

    try:
        os.makedirs(opt.outdir)
    except OSError:
        pass

    train_writer = SummaryWriter(os.path.join(log_dirname, 'train'))
    test_writer = SummaryWriter(os.path.join(log_dirname, 'test'))
    optimizer = torch.optim.Adam(cnnnet.parameters(), lr=opt.lr)

    # milestones in number of optimizer iterations
    cnnnet.to(device)
    train_num_batch = len(train_dataloader)
    test_num_batch = len(test_dataloader)

    global lsr_matrix
    lsr_matrix = lsr_matrix.unsqueeze(0).expand(opt.batchSize,-1,-1).to(torch.float64).to(device)
    
    start = time.time()
    for epoch in range(opt.nepoch):
        train_batchind = -1
        train_fraction_done = 0.0
        train_enum = enumerate(train_dataloader, 0)

        test_batchind = -1
        test_fraction_done = 0.0
        test_enum = enumerate(test_dataloader, 0)
        
        for train_batchind, data in train_enum:
            cnnnet.eval()
            
            # construct feats
            input = data[0]
            input_lsr = torch.zeros(27,opt.batchSize,w,w,w)
            count = 0
            for x in range(3):
                for y in range(3):
                    for z in range(3):
                        input_lsr[count,:,:,:,:] = input[:,0,x:x+w,y:y+w,z:z+w]
                        count+=1
            input_lsr = input_lsr.permute(1,0,2,3,4) 
            input_lsr = torch.reshape(input_lsr, (opt.batchSize*27,w,w,w))
            input = input_lsr.unsqueeze(1)
            input = input.to(device)            
                                           
            target = data[1]
            target = target.to(device)

            # forward pass
            input.requires_grad = True
            pred = cnnnet(input)
            pred = torch.reshape(pred, (opt.batchSize,27,1))

            # compute loss
            loss = compute_loss(
                pred=pred, target=target,
                device = device)

            # zero gradients
            optimizer.zero_grad()
            
            # backward pass 
            loss.backward()

            # parameter optimization step
            optimizer.step()

            # print info and update log file
            print('[%s %d: %d/%d] %s loss: %f' % (opt.name, epoch,
                  train_batchind, train_num_batch-1, green('train'), loss.item()))
            train_writer.add_scalar('loss', loss.item(
            ), (epoch + train_fraction_done) * train_num_batch * opt.batchSize)
            train_fraction_done = (train_batchind+1) / train_num_batch
            
            while test_fraction_done <= train_fraction_done and test_batchind+1 < test_num_batch:
                # set to evaluation mode
                cnnnet.eval()
                test_batchind, data = next(test_enum)

                input = data[0]
                input_lsr = torch.zeros(27,opt.batchSize,w,w,w)
                count = 0
                for x in range(3):
                    for y in range(3):
                        for z in range(3):
                            input_lsr[count,:,:,:,:] = input[:,0,x:x+w,y:y+w,z:z+w]
                            count+=1
                input_lsr = input_lsr.permute(1,0,2,3,4) 
                input_lsr = torch.reshape(input_lsr, (opt.batchSize*27,w,w,w))
                input = input_lsr.unsqueeze(1)
                input = input.to(device)                                 
            
                target = data[1]
                target = target.to(device)
                    
                # forward pass
                pred = cnnnet(input)
                pred = torch.reshape(pred, (opt.batchSize,27,1))

                # compute loss
                loss = compute_loss(
                    pred=pred, target=target,
                    device = device)

                # print info and update log file
                print('[%s %d: %d/%d] %s loss: %f' % (opt.name, epoch,
                      train_batchind, train_num_batch-1, blue('test'), loss.item()))
                test_writer.add_scalar('loss', loss.item(
                ), (epoch + test_fraction_done) * train_num_batch * opt.batchSize)

                test_fraction_done = (test_batchind+1) / test_num_batch

        # save model, overwriting the old model
        if epoch % opt.saveinterval == 0 or epoch == opt.nepoch-1:
            torch.save(cnnnet.state_dict(), model_filename)

    print("time used is: %.2f" %(time.time()-start))  


 
def compute_loss(pred, target, device):
    # mean squared error loss term 
    loss1 = torch.mean((pred - target) ** 2)
    pred_sq = pred.squeeze().unsqueeze(1).to(torch.float64)

    # polynomial loss term
    loss2 = torch.mean(torch.sum(torch.einsum('bij,bjk->bik',pred_sq, lsr_matrix).squeeze()*pred.squeeze(),dim=1))
    
    loss = loss1 + loss2 
    
    print('loss_p %f, loss_g %f' %(loss1, loss2))
    return loss


if __name__ == '__main__':
    train_opt = parse_arguments()
    train_cnnnet(train_opt)
