#!/bin/bash

GENERATE_DATA=1 # true = do data generation
DO_SELECT=1 # true = select data from generate data for training
TRAIN_MODEL=1 # true = do model training

function usage {
    if [ "$#" -ne 1 ]; then
        echo "$#"
        echo "Usage: $0 <model>"
        exit 1
    fi
}

usage "$@"
model_name=$1

#--------------------------------------------------------------------------------
# parameters for generating the training data
#--------------------------------------------------------------------------------
# types of shapes to use for training
# (shapes available: sphere, cone, torus, cylinder, oriented_box)
SHAPES="sphere cone torus cylinder"
# size of training data subsets
# (there are 9 subsets)
N=40000
# particles per cell used in training
PPC_half=0.5
PPC1=1
PPC2=2
# minimum allowed distance between a particle and the surface
gap=0.25
# thickness (cells) for thin sheets
sheet_thickness=0.5
# true = include thin sheets in training data
do_thin_sheet=1
# true = include inverse shapes in training data
do_inv_shape=1
# training/validation data split
percent_training=.95
percent_validation=.05
#--------------------------------------------------------------------------------

# number of processes to run in parallel
P=20
base=shapes
function generate_shapes_analytic {
    while [[ $(cat 2>/dev/null $base/$1.*.m0 | wc -l) -le $((N * $7)) ]]; do
        i=$(ls 2>/dev/null $base/$1.*.grid.gz | wc -l)
        for j in $( seq 1 $P ) ; do
            echo $1 $i $2 $3 $4 $5 $6;
            i=$(( i+1 ));
        done  | xargs -P $P -n 7  ./generate_shapes.sh $base
    done
}

if [ $GENERATE_DATA = 1 ] ; then
    rm -rf $base
    mkdir $base

    echo "Generating training data..."
    
    # inverse shapes
    for s in $SHAPES
    do
        generate_shapes_analytic $s $PPC1 0 $sheet_thickness $do_inv_shape $gap 1
        generate_shapes_analytic $s $PPC1 0 $sheet_thickness $do_inv_shape $gap 2
        generate_shapes_analytic $s $PPC2 0 $sheet_thickness $do_inv_shape $gap 3
    done
    
    # regular shapes
    for s in $SHAPES
    do
        generate_shapes_analytic $s $PPC1 0 $sheet_thickness 0 $gap 4       
        generate_shapes_analytic $s $PPC1 0 $sheet_thickness 0 $gap 5
        generate_shapes_analytic $s $PPC2 0 $sheet_thickness 0 $gap 6
    done

    # thin sheets
    for s in $SHAPES
    do
        generate_shapes_analytic $s $PPC_half $do_thin_sheet $sheet_thickness 0 $gap 7
        generate_shapes_analytic $s $PPC_half $do_thin_sheet $sheet_thickness 0 $gap 8
        generate_shapes_analytic $s $PPC1 $do_thin_sheet $sheet_thickness 0 $gap 9
    done
        
    echo "GENERATE SHAPES FINISHED"
fi
#--------------------------------------------------------------------------------
# select data to use and separate into training and validation subsets
#--------------------------------------------------------------------------------
T1=$(echo "$N*9*$percent_training" | bc) # training data
V1=$(echo "$N*9*$percent_validation" | bc) # validation data

base_dataset=cnn_dataset
if [ $DO_SELECT == 1 ] ; then

    rm -rf $base_dataset
    mkdir $base_dataset
    
    rest=""
    for s in $SHAPES
    do
        rest="$rest -p $base/$s.%d -t $T1 -v $V1 -shape $s"
    done
    
    train="training"
    valid="validation"

    ./select_larger_training_data \
        -t_prefix $base_dataset/$train \
        -v_prefix $base_dataset/$valid \
        -t_names $base_dataset/trainingset.txt \
        -v_names $base_dataset/validationset.txt \
        -e '.m0' \
        -e '.depth'  \
        $rest
    
    # convert training data from ascii to numpy
    rm -f all.txt
    cat $base_dataset/*.txt > $base_dataset/all.txt
    while IFS= read -r line; do
        echo $line
    done < $base_dataset/all.txt \
        | xargs -P 40 -I {} \
                python3 ./save_as_numpy.py \
                --indir $base_dataset \
                --shape_names {} \
                --inputs ".m0" \
                --outputs ".depth"\
                --width "10"

fi
#--------------------------------------------------------------------------------
# train the model
#--------------------------------------------------------------------------------
# run python code (train)
if [ $TRAIN_MODEL == 1 ]; then
    train_set=""
    test_set=""
    
    while IFS= read -r line; do
        train_set="$train_set$line "
    done < $base_dataset/trainingset.txt

    while IFS= read -r line; do
        test_set="$test_set$line "
    done < $base_dataset/validationset.txt    
    
    python3 ./train_cnnnet.py --name $model_name \
            --indir $base_dataset \
            --trainset $train_set \
            --testset $test_set \
            --batchSize 128 \
            --nepoch 5 \
                             
    echo TRAINING FINISHED
fi
